import { DancerIdentity, Rotation, normalizeRotation } from "./danceCommon.js";
import { DancerSetPosition, DancersSetPositions, Hand } from "./rendererConstants.js";
import { Offset, leftShoulder, rightShoulder, degreesToRadians, radiansToDegrees } from "./rendererConstants.js";

export enum AnimationKind {
  Linear = "Linear",
  Rotation = "Rotation",
}

export interface RotationAround {
  center: Offset,
  width: number,
  height: number,
};

export abstract class AnimationSegment {
  public readonly beats: number;
  // Start position is redundant... but allows AnimationSegments to be defined indepent of their surroundings.
  public readonly startPosition: DancerSetPosition;
  public readonly endPosition: DancerSetPosition;

  protected constructor(beats: number, startPosition: DancerSetPosition, endPosition: DancerSetPosition) {
    this.beats = beats;
    this.startPosition = startPosition;
    this.endPosition = endPosition;
  }

  abstract interpolateOffset(progress: number);

  // TODO Not sure this interpolation is actually right for arms...
  interpolateHandOffset(progress: number, hand: Hand) : Offset | undefined {
    let prev = hand === Hand.Left ? this.startPosition.leftArmEnd : this.startPosition.rightArmEnd;
    let next = hand === Hand.Left ? this.endPosition.leftArmEnd : this.endPosition.rightArmEnd;

    if (!prev && !next) return undefined;
    prev ??= hand.shoulderPosition();
    next ??= hand.shoulderPosition();

    return interpolateLinearOffset(progress, prev, next);
  }

  interpolateRotation(progress: number) {
    return interpolateLinear(progress, this.startPosition.rotation, this.endPosition.rotation);
  }

  positionAtFraction(progress: number) : DancerSetPosition {
    return {
      position: this.interpolateOffset(progress),
      rotation: this.interpolateRotation(progress),
      leftArmEnd: this.interpolateHandOffset(progress, Hand.Left),
      rightArmEnd: this.interpolateHandOffset(progress, Hand.Right),
    }
  }
}

function interpolateLinear(progress: number, prev: number, next: number): number {
  return next * progress + prev * (1 - progress);
}

function interpolateLinearOffset(progress: number, prev: Offset, next: Offset): Offset {
  return {
    x: interpolateLinear(progress, prev.x, next.x),
    y: interpolateLinear(progress, prev.y, next.y)
  };
}

export class LinearAnimationSegment extends AnimationSegment {
  constructor(beats: number, startPosition: DancerSetPosition, endPosition: DancerSetPosition) {
    super(beats, startPosition, endPosition);
  }

  public static standStill(startAndEndPosition: DancerSetPosition, beats?: number) {
    return new LinearAnimationSegment(beats ?? 0, startAndEndPosition, startAndEndPosition);
  }

  override interpolateOffset(progress: number) {
    return interpolateLinearOffset(progress, this.startPosition.position, this.endPosition.position);
  }
}

export interface AnimationTransitionFlags {
  rotation?: boolean;
  hands?: boolean;
}
export class TransitionAnimationSegment extends AnimationSegment {
  private readonly actualAnimation: AnimationSegment;
  private readonly flags: AnimationTransitionFlags;
  private readonly startTransitionProgress: number;
  private readonly endTransitionProgress: number;
  private readonly startRotation: number;
  private readonly endRotation: number;

  constructor({ actualAnimation, flags, startTransitionBeats, endTransitionBeats }: { actualAnimation: AnimationSegment; flags: AnimationTransitionFlags; startTransitionBeats: number; endTransitionBeats?: number; }) {
    super(actualAnimation.beats, actualAnimation.startPosition, actualAnimation.endPosition);

    this.actualAnimation = actualAnimation;
    this.flags = flags;
    this.startTransitionProgress = startTransitionBeats / actualAnimation.beats;
    this.endTransitionProgress = endTransitionBeats === undefined ? this.startTransitionProgress : endTransitionBeats / actualAnimation.beats;

    this.startRotation = this.startPosition.rotation;
    this.endRotation = this.endPosition.rotation;
    if (this.flags.rotation) {
      const actualStart = this.actualAnimation.interpolateRotation(0);
      const actualEnd = this.actualAnimation.interpolateRotation(1);

      const rotationPositive = actualEnd > actualStart;

      const transitionStart = this.actualAnimation.interpolateRotation(this.startTransitionProgress);
      const transitionEnd = this.actualAnimation.interpolateRotation(1 - this.endTransitionProgress);

      if (rotationPositive) {
        while (transitionStart < this.startRotation) {
          this.startRotation -= 360;
        }
        while (transitionEnd > this.endRotation) {
          this.endRotation += 360;
        }
      } else {
        while (transitionStart > this.startRotation) {
          this.startRotation += 360;
        }
        while (transitionEnd < this.endRotation) {
          this.endRotation -= 360;
        }
      }
    }
  }

  override interpolateOffset(progress: number) {
    return this.actualAnimation.interpolateOffset(progress);
  }

  override interpolateRotation(progress: number): number {
    const actualRotation = this.actualAnimation.interpolateRotation(progress);

    if (this.flags.rotation) {
      if (progress < this.startTransitionProgress) {
        return interpolateLinear(progress / this.startTransitionProgress, this.startRotation, actualRotation);
      } else if ((1 - progress) < this.endTransitionProgress) {
        return interpolateLinear((1 - progress) / this.endTransitionProgress, this.endRotation, actualRotation);
      }
    }

    return actualRotation;
  }

  override interpolateHandOffset(progress: number, hand: Hand): Offset | undefined {
    const actualHandOffset = this.actualAnimation.interpolateHandOffset(progress, hand);

    if (this.flags.hands) {
      if (progress < this.startTransitionProgress) {
        const startHand = hand === Hand.Left
          ? this.startPosition.leftArmEnd ?? leftShoulder
          : this.startPosition.rightArmEnd ?? rightShoulder;
        return interpolateLinearOffset(progress / this.startTransitionProgress, startHand, actualHandOffset ?? hand.shoulderPosition());
      } else if (1 - progress < this.endTransitionProgress) {
        const endHand = hand === Hand.Left
          ? this.endPosition.leftArmEnd ?? leftShoulder
          : this.endPosition.rightArmEnd ?? rightShoulder;
        return interpolateLinearOffset((1 - progress) / this.endTransitionProgress, endHand, actualHandOffset ?? hand.shoulderPosition());
      }
    }

    return actualHandOffset;
  }
}

// TODO Not sure this really belongs here instead of on some Semantic* type.
export enum RotationAnimationFacing {
  Linear = "Linear", // Default, linearly interpolate.
  Center = "Center", // Always face the center.
  CenterRelative = "CenterRelative", // Always face the center.
  Forward = "Forward", // Always face the direction of the rotation.
  Backward = "Backward", // Opposite of forward.
}
export interface CloserDuringRotation {
  minDistance: number,
  transitionBeats: number,
}
interface Closer {
  middleDistance: number,
  transitionProgress: number,
}
export interface HandAnimation {
  kind: "Linear" | "Center" | "None" | "Start" | "End"
}

export class RotationAnimationSegment extends AnimationSegment {
  private readonly xRadius: number;
  private readonly yRadius: number;
  private readonly startDistance: number;
  private readonly endDistance: number;
  private readonly startRotation: number;
  private readonly endRotation: number;
  private readonly center: Offset;
  private readonly facing: RotationAnimationFacing;
  private readonly startFacing: Rotation;
  private readonly closer?: Closer;
  private readonly hands: Map<Hand, HandAnimation>;

  constructor(beats: number, startPosition: DancerSetPosition, endPosition: DancerSetPosition,
              rotation: number, around: RotationAround, facing: RotationAnimationFacing,
              closer?: CloserDuringRotation, hands?: Map<Hand, HandAnimation>) {
    super(beats, startPosition, endPosition);

    this.hands = new Map<Hand, HandAnimation>(
      [Hand.Left, Hand.Right].map(h => [h, hands?.get(h) ?? { kind: "Linear" }]));

    this.facing = facing;
    this.xRadius = around.width / 2;
    this.yRadius = around.height / 2;
    this.center = around.center;
    let positionToDegrees = (pos: Offset): number => {
      const radians = Math.atan2(
        (pos.y - around.center.y) / this.yRadius,
        (pos.x - around.center.x) / this.xRadius
      );
      // Canvas rotation is backwards of what we expect, so take the negative here.
      return -radiansToDegrees(radians);
    }
    this.startRotation = positionToDegrees(startPosition.position);
    this.startFacing = startPosition.rotation;
    const actualRotation = normalizeRotation(positionToDegrees(endPosition.position) - this.startRotation,
      rotation);
    this.endRotation = this.startRotation + actualRotation;

    let positionToNormalizedDistance = (pos: Offset): number => {
      return Math.sqrt(Math.pow((pos.y - around.center.y) / this.yRadius, 2)
                     + Math.pow((pos.x - around.center.x) / this.xRadius, 2));
    }
    this.startDistance = positionToNormalizedDistance(startPosition.position);
    this.endDistance = positionToNormalizedDistance(endPosition.position);
    if (closer && (this.startDistance > closer.minDistance || this.endDistance > closer.minDistance)) {
      this.closer = {
        middleDistance: closer.minDistance,
        transitionProgress: closer.transitionBeats / beats,
      };
    }
  }

  override interpolateOffset(progress: number) {
    const radians = -degreesToRadians(interpolateLinear(progress, this.startRotation, this.endRotation));
    let distance: number;
    if (!this.closer) {
      distance = interpolateLinear(progress, this.startDistance, this.endDistance);
    } else {
      if (progress <= this.closer.transitionProgress) {
        distance = interpolateLinear(progress / this.closer.transitionProgress, this.startDistance, this.closer.middleDistance);
      } else if ((1 - progress) <= this.closer.transitionProgress) {
        distance = interpolateLinear((1 - progress) / this.closer.transitionProgress, this.endDistance, this.closer.middleDistance);
      } else {
        distance = this.closer.middleDistance;
      }
    }
    return {
      x: this.center.x + this.xRadius * Math.cos(radians) * distance,
      y: this.center.y + this.yRadius * Math.sin(radians) * distance,
    }
  }

  override interpolateRotation(progress: number): number {
    if (this.facing === RotationAnimationFacing.Linear) {
      return super.interpolateRotation(progress);
    } else {
      const degrees = interpolateLinear(progress, this.startRotation, this.endRotation);
      let rotation : number;
      switch (this.facing) {
        case RotationAnimationFacing.Center:
          return degrees - 90;
        case RotationAnimationFacing.CenterRelative:
          return degrees - this.startRotation + this.startFacing;
        case RotationAnimationFacing.Forward:
          return degrees + 180;
        case RotationAnimationFacing.Backward:
          return degrees;
      }
    }
  }

  override interpolateHandOffset(progress: number, hand: Hand): Offset | undefined {
    const handAnim : HandAnimation = this.hands.get(hand)!;
    let middlePos : Offset;
    switch (handAnim.kind) {
      case "Start":
        return hand === Hand.Left ? this.startPosition.leftArmEnd : this.startPosition.rightArmEnd;
      case "End":
        return hand === Hand.Left ? this.endPosition.leftArmEnd : this.endPosition.rightArmEnd;
      case "Linear":
        return super.interpolateHandOffset(progress, hand);
      case "Center":
        const position = this.interpolateOffset(progress);
        const offset = {
          x: position.x - this.center.x,
          y: position.y - this.center.y,
        };
        const rotation = degreesToRadians(this.interpolateRotation(progress));

        const centerPos = {
          x: -Math.cos(rotation) * offset.x + Math.sin(rotation) * offset.y,
          y: -Math.sin(rotation) * offset.x - Math.cos(rotation) * offset.y,
        }
        return centerPos;
      case "None":
        return hand === Hand.Left ? leftShoulder : rightShoulder;
    }
  }
}

export type Animation = Map<DancerIdentity, AnimationSegment[]>;
export function newAnimation() : Animation { return new Map<DancerIdentity, AnimationSegment[]>() };

function positionAtBeat(animation: AnimationSegment[], beat: number) : DancerSetPosition {
  let lastFrame : AnimationSegment = animation[0];
  let lastFrameEndBeat = 0;
  for (const frame of animation) {
    let currentFrameEndBeat = lastFrameEndBeat + frame.beats;
    if (currentFrameEndBeat < beat) {
      lastFrame = frame;
      lastFrameEndBeat = currentFrameEndBeat;
      continue;
    } else if (currentFrameEndBeat === beat) {
      return frame.endPosition;
    } else {
      const progress = (beat - lastFrameEndBeat) / (currentFrameEndBeat - lastFrameEndBeat);
      return frame.positionAtFraction(progress);
    }
  }

  // fallthrough
  return lastFrame.endPosition;
}

export function positionsAtBeat(animation: Animation, beat: number) : DancersSetPositions {
    const res = new Map<DancerIdentity, DancerSetPosition>();
    for (const [id, animationSegments] of animation.entries()) {
      res.set(id, positionAtBeat(animationSegments, beat));
    }
    return res;
}