import { Rotation } from "./danceCommon";
import { Hand } from "./rendererConstants";

export enum PositionKind {
  Circle = "Circle",
  ShortLines = "ShortLines",
};
// Which of the four corners of a circle.
// Note that "Left" is the left side of the set, not left from the dancer's PoV,
//  so, in improper, the Robin starts in the BottomLeft.
enum CirclePositionEnum {
  TopLeft = "TopLeft",
  BottomLeft = "BottomLeft",
  BottomRight = "BottomRight",
  TopRight = "TopRight",
}
export enum CircleSide {
  Top = "Top",
  Bottom = "Bottom",
  Left = "Left",
  Right = "Right",
}
export type CircleSideOrCenter = CircleSide | "Center";
export class CirclePosition {
  public static readonly TopLeft = new CirclePosition(CirclePositionEnum.TopLeft);
  public static readonly BottomLeft = new CirclePosition(CirclePositionEnum.BottomLeft);
  public static readonly BottomRight = new CirclePosition(CirclePositionEnum.BottomRight);
  public static readonly TopRight = new CirclePosition(CirclePositionEnum.TopRight);

  private readonly enumValue: CirclePositionEnum;

  private constructor(enumValue: CirclePositionEnum) {
    this.enumValue = enumValue;
  }

  private static enumValueToNumber(enumValue: CirclePositionEnum) : number {
    switch (enumValue) {
      case CirclePositionEnum.TopLeft:
        return 0;
      case CirclePositionEnum.BottomLeft:
        return 1;
      case CirclePositionEnum.BottomRight:
        return 2;
      case CirclePositionEnum.TopRight:
        return 3;
    }
  }

  private static numberToEnumValue(num: number) : CirclePositionEnum {
    if (num < 0 || num > 3) {
      num %= 4;
      if (num < 0) num += 4;
    }

    return [
      CirclePositionEnum.TopLeft,
      CirclePositionEnum.BottomLeft,
      CirclePositionEnum.BottomRight,
      CirclePositionEnum.TopRight,
    ][num];
  }

  private static get(enumValue: CirclePositionEnum) : CirclePosition {
    return [
      CirclePosition.TopLeft,
      CirclePosition.BottomLeft,
      CirclePosition.BottomRight,
      CirclePosition.TopRight,
    ][this.enumValueToNumber(enumValue)];
  }

  public circleRight(places?: number) : CirclePosition {
    return CirclePosition.get(CirclePosition.numberToEnumValue(CirclePosition.enumValueToNumber(this.enumValue) - (places ?? 1)));
  }

  public circleLeft(places?: number) : CirclePosition {
    return CirclePosition.get(CirclePosition.numberToEnumValue(CirclePosition.enumValueToNumber(this.enumValue) + (places ?? 1)));
  }

  public swapAcross() : CirclePosition {
    return [
      CirclePosition.TopRight,
      CirclePosition.BottomRight,
      CirclePosition.BottomLeft,
      CirclePosition.TopLeft,
    ][CirclePosition.enumValueToNumber(this.enumValue)];
  }

  public swapUpAndDown() : CirclePosition {
    return [
      CirclePosition.BottomLeft,
      CirclePosition.TopLeft,
      CirclePosition.TopRight,
      CirclePosition.BottomRight,
    ][CirclePosition.enumValueToNumber(this.enumValue)];
  }

  public facingCenterRotation() : Rotation {
    return (45 + 90 * CirclePosition.enumValueToNumber(this.enumValue)) % 360;
  }

  public topBottomSide() : CircleSide.Top | CircleSide.Bottom {
    return this.enumValue === CirclePositionEnum.TopLeft || this.enumValue === CirclePositionEnum.TopRight
      ? CircleSide.Top
      : CircleSide.Bottom;
  }

  public leftRightSide() : CircleSide.Left | CircleSide.Right {
    return this.enumValue === CirclePositionEnum.TopLeft || this.enumValue === CirclePositionEnum.BottomLeft
      ? CircleSide.Left
      : CircleSide.Right;
  }

  public isTop() : boolean {
    return this.topBottomSide() === CircleSide.Top;
  }

  public isLeft() : boolean {
    return this.leftRightSide() === CircleSide.Left;
  }

  public isOnLeftLookingAcross() : boolean {
    return this.enumValue === CirclePositionEnum.TopRight || this.enumValue === CirclePositionEnum.BottomLeft;
  }

  public isOnLeftLookingUpAndDown() : boolean {
    return this.enumValue === CirclePositionEnum.TopLeft || this.enumValue === CirclePositionEnum.BottomRight;
  }

  public toString() : string {
    return this.enumValue.toString();
  }
}

export enum Facing {
  CenterOfCircle = "CenterOfCircle",
  Up = "Up",
  Down = "Down",
  Left = "Left",
  Right = "Right",
}
export enum HandTo {
  LeftInCircle = "LeftInCircle",
  RightInCircle = "RightInCircle",
  AcrossCircle = "AcrossCircle",
  DancerForward = "DancerForward",
  DancerLeft = "DancerLeft",
  DancerRight = "DancerRight",
}
export interface HandConnection {
  to: HandTo,
  hand: Hand,
}
export enum BalanceWeight {
  Forward = "Forward",
  Backward = "Backward",
  Left = "Left",
  Right = "Right",
}
export enum DancerDistance {
  Normal = "Normal",
  Compact = "Compact", // allemande, etc.
  // Swings asymmetrical, but some dances may have the Lark do a swing as a Robin or vice versa,
  //  especially any dance with Larks Swing or Robins Swing.
  SwingLark = "SwingLark",
  SwingRobin = "SwingRobin",
}
export type SemanticPosition = {
  kind: PositionKind.Circle,
  setOffset?: number,
  lineOffset?: number,
  which: CirclePosition,
  facing: Facing,
  hands?: Map<Hand, HandConnection>,
  balance?: BalanceWeight,
  dancerDistance?: DancerDistance,
} | {
  kind: PositionKind.ShortLines,
  setOffset?: number,
  lineOffset?: number,
  hands?: Map<Hand, HandConnection>,
  balance?: BalanceWeight,
};
