import * as animation from "./animation.js";
import * as common from "./danceCommon.js";
import { DanceRole, DancerIdentity, Rotation } from "./danceCommon.js";
import { BalanceWeight, CirclePosition, CircleSide, CircleSideOrCenter, DancerDistance, Facing, HandConnection, HandTo, PositionKind, SemanticPosition } from "./interpreterCommon.js";
import { Move } from "./libfigureMapper.js";
import { DancerSetPosition, Hand, Offset, dancerWidth, lineDistance, setDistance, setHeight, setSpacing, setWidth } from "./rendererConstants.js";


export enum SemanticAnimationKind {
  StandStill = "StandStill",

  // Long lines, balances?, slide, walk forward to new wave
  Linear = "Linear",

  // Rotate around a set: circle left/right, star
  Circle = "Circle",

  // Walk in rectangle pattern passing by someone. Also Mad Robin and See Saw. Pousset might fit here?
  DoSiDo = "DoSiDo",

  // Also pull by.
  PassBy = "PassBy",

  // California twirl, box the gnat, etc.
  TwirlSwap = "TwirlSwap",

  // Rotate as a pair around a point. Allemande, right/left shoulder round
  RotateAround = "RotateAround",

  Swing = "Swing",
}
export type SemanticAnimation = {
  kind: SemanticAnimationKind.StandStill,
} | {
  kind: SemanticAnimationKind.Linear,

  // Approximate rotation. endPosition determines final orientation, generally a multiple of +/-360.
  minRotation?: number,
} | {
  kind: SemanticAnimationKind.Circle,

  minRotation?: number,

  // positive is to the left.
  places: number,
} | {
  kind: SemanticAnimationKind.DoSiDo,

  minRotation?: number,

  // amount to do si do, measured in degrees clockwise (negative for counterclockwise/passing by left)
  amount: number,
} | {
  kind: SemanticAnimationKind.RotateAround,

  minAmount: number,

  around: CircleSide | "Center",

  byHand: Hand | undefined,
} | {
  kind: SemanticAnimationKind.Swing,

  minAmount: number,

  around: CircleSide | "Center",

  endFacing: Facing,

  // Swings are asymmetric. This is usually but not always the dancer's role.
  swingRole: common.DanceRole,
} | {
  kind: SemanticAnimationKind.TwirlSwap,

  around: CircleSide,
}


// TODO
export interface LowLevelMove {
  // for debugging messages
  remarks?: string,
  // move is here for reference/debugging, shouldn't actually get used.
  move: Move,
  // for debugging, if move has been broken into multiple low-level moves, where does this one start
  startBeat: number,

  // How many beats does this take up?
  beats: number,

  // TODO decide if startPosition, endPosition, or both need to adjust for adjacent moves.
  // TODO when are hands optional?
  // Position this should start in. Previous move might not end here? Or should adjust to end here?
  startPosition: SemanticPosition,
  // Position this should end in. Next move night not star there? Or should adjust to start here?
  endPosition: SemanticPosition,

  movementPattern: SemanticAnimation,
  // TODO assertion of spin direction?
}


function CenterOfSet(setOffset?: number, lineOffset?: number): Offset {
  let position = {x: 0, y: 0};

  if (setOffset) position.y += setOffset * setDistance;
  if (lineOffset) position.x += lineOffset * lineDistance;

  return position;
}
function CenterOf(sideOrCenter: CircleSideOrCenter, setOffset?: number, lineOffset?: number): Offset {
  const setCenter = CenterOfSet(setOffset, lineOffset);

  if (sideOrCenter === "Center") return setCenter;
  const side : CircleSide = sideOrCenter;

  const xOffset = side === CircleSide.Left ? -1 : side === CircleSide.Right ? +1 : 0;
  const yOffset = side === CircleSide.Top ? -1 : side === CircleSide.Bottom ? +1 : 0;

  return { x: setCenter.x + xOffset, y: setCenter.y + yOffset };
}

function SemanticToSetPosition(semantic: SemanticPosition): DancerSetPosition {
  switch (semantic.kind) {
    case PositionKind.Circle:
      let rotation: number;
      switch (semantic.facing) {
        case Facing.CenterOfCircle:
          rotation = semantic.which.facingCenterRotation();
          break;
        case Facing.Up:
          rotation = Rotation.Up;
          break;
        case Facing.Down:
          rotation = Rotation.Down;
          break;
        case Facing.Left:
          rotation = Rotation.Left;
          break;
        case Facing.Right:
          rotation = Rotation.Right;
          break;
        default:
          throw "Unexpected facing: " + semantic.facing;
      }
      let yAmount : number;
      switch (semantic.dancerDistance) {
        case DancerDistance.Normal:
        case undefined:
          yAmount = 1;
          break;
        case DancerDistance.Compact:
          yAmount = 0.75;
          break;
        case DancerDistance.SwingLark:
        case DancerDistance.SwingRobin:
          yAmount = 0.5;
          break;
        default:
          throw "Unknown DancerDistance: " + semantic.dancerDistance;
      }
      let position: Offset;
      switch (semantic.which) {
        case CirclePosition.TopLeft:
          position = { x: -1, y: -yAmount };
          break;
        case CirclePosition.TopRight:
          position = { x: +1, y: -yAmount };
          break;
        case CirclePosition.BottomLeft:
          position = { x: -1, y: +yAmount };
          break;
        case CirclePosition.BottomRight:
          position = { x: +1, y: +yAmount };
          break;
        default:
          throw "Invalid circle position: " + semantic.which;
      }
      const balanceAmount = 0.25;
      if (semantic.balance) {
        if (semantic.facing === Facing.CenterOfCircle) {
          if (semantic.balance === BalanceWeight.Forward) {
            position = { x: position.x * (1-balanceAmount), y: position.y * (1-balanceAmount) };
          } else {
            throw "Don't know what balancing in a circle not forward means. BalanceWeight=" + semantic.balance;
          }
        } else if (semantic.facing === Facing.Up || semantic.facing === Facing.Down
                && (semantic.balance === BalanceWeight.Backward || semantic.balance === BalanceWeight.Forward)) {
          position = {
            x: position.x,
            y: position.y + position.y * (
              semantic.balance === BalanceWeight.Forward
              ? -balanceAmount
              : balanceAmount)
          };
        }
      }
      const setOffset = CenterOfSet(semantic.setOffset, semantic.lineOffset);
      position.x += setOffset.x;
      position.y += setOffset.y;

      const isFacingUpOrDown = semantic.facing === Facing.Up || semantic.facing === Facing.Down;
      if (semantic.dancerDistance === DancerDistance.SwingLark) {
        if (isFacingUpOrDown) {
          rotation -= 45;
        } else {
          rotation += 45;
        }

        return {
          position,
          rotation,
          leftArmEnd: { x: 0, y: Math.sqrt(1/2)},
          rightArmEnd: { x: 0.8, y: 0.5 },
        };
      } else if (semantic.dancerDistance === DancerDistance.SwingRobin) {
        if (isFacingUpOrDown) {
          rotation += 45;
        } else {
          rotation -= 45;
        }

        return {
          position,
          rotation,
          leftArmEnd: { x: -0.8, y: 0.5 },
          rightArmEnd: { x: 0, y: Math.sqrt(1/2) },
        };
      }

      function processHand(hand: Hand, connection: HandConnection | undefined): Offset | undefined {
        // TODO Hands. Might need more info? Or need context of nearby dancer SemanticPositions?
        if (!connection) return undefined;

        switch (connection.to) {
          case HandTo.DancerLeft:
            return { x: -1, y: 0 };
          case HandTo.DancerRight:
            return { x: +1, y: 0 };
          case HandTo.DancerForward:
            if (hand === connection.hand) {
              return { x: 0, y: +0.5 };
            } else {
              return { x: dancerWidth / 2 * (hand === Hand.Left ? -1 : +1), y: +1 };
            }
          case HandTo.LeftInCircle:
          case HandTo.RightInCircle:
            const xSign = connection.to === HandTo.LeftInCircle ? -1 : +1;
            const balanceFactor = semantic.balance === BalanceWeight.Forward ? 1-balanceAmount : 1;
            return { x: balanceFactor*xSign/Math.sqrt(2), y: balanceFactor/Math.sqrt(2) };
          default:
            throw "Unkown connection: " + connection.to;
        }
      }


      return {
        position,
        rotation,
        leftArmEnd: processHand(Hand.Left, semantic.hands?.get(Hand.Left)),
        rightArmEnd: processHand(Hand.Right, semantic.hands?.get(Hand.Right)),
      };
    default:
      throw "Unsupported PositionKind: " + semantic.kind;
  }
}

export function animateLowLevelMove(move: LowLevelMove): animation.AnimationSegment[] {
  const startSetPosition = SemanticToSetPosition(move.startPosition);
  const endSetPosition = SemanticToSetPosition(move.endPosition);

  function handleMove(): animation.AnimationSegment[] {
    switch (move.movementPattern.kind) {
      case SemanticAnimationKind.StandStill:
        // TODO Where is the check that the current position is already move.endPosition?
        //      And that end and start position are the same... modulo hands and rotation modulus?
        return [
          animation.LinearAnimationSegment.standStill(startSetPosition, move.beats),
        ];
      case SemanticAnimationKind.Linear:
        let rotation = endSetPosition.rotation - startSetPosition.rotation;
        const minRotation = move.movementPattern.minRotation;
        try {
          rotation = common.normalizeRotation(rotation, minRotation);
        } catch {
          throw "Expected zero rotation, but start and end positions at different orientations: start="
          + JSON.stringify(move.startPosition) + ", end=" + JSON.stringify(move.endPosition);
        }
        return [
          new animation.LinearAnimationSegment(move.beats,
            startSetPosition,
            {
              ...endSetPosition,
              rotation: startSetPosition.rotation + rotation,
            }),
        ];
      case SemanticAnimationKind.Circle:
        if (move.startPosition.kind !== PositionKind.Circle || move.endPosition.kind !== PositionKind.Circle) {
          throw "Circle must start and end in a circle.";
        }

        return [
          new animation.TransitionAnimationSegment({
            actualAnimation: new animation.RotationAnimationSegment(move.beats,
              startSetPosition,
              endSetPosition,
              move.movementPattern.places * 90,
              {
                center: CenterOfSet(move.startPosition.setOffset, move.startPosition.lineOffset),
                width: setWidth,
                height: setHeight,
              },
              animation.RotationAnimationFacing.Center,
              undefined,
              new Map<Hand, animation.HandAnimation>([
                [Hand.Left, { kind: "End" }],
                [Hand.Right, { kind: "End" }],
              ])
            ),
            flags: {
              hands: true
            },
            startTransitionBeats: 1
          }),
        ];
      case SemanticAnimationKind.DoSiDo:
        // TODO Rotation
        // TODO Parameters: rotation, amount, direction, diagonal?

        if (move.startPosition.kind !== PositionKind.Circle) {
          throw "DoSiDo must start in a circle: " + JSON.stringify(move);
        }
        const center = CenterOf(move.startPosition.which.leftRightSide(), move.startPosition.setOffset, move.startPosition.lineOffset);
        return [
          new animation.RotationAnimationSegment(move.beats,
            startSetPosition,
            endSetPosition,
            move.movementPattern.amount,
            {
              center,
              width: setWidth / 3,
              height: setHeight,
            },
            animation.RotationAnimationFacing.Linear,
          ),
        ];
      case SemanticAnimationKind.RotateAround:
        // TODO Rotation
        // TODO Parameters: rotation, amount, direction, diagonal?

        if (move.startPosition.kind !== PositionKind.Circle) {
          throw "Rotate must start in a circle: " + JSON.stringify(move);
        }
        // TODO Could rotate around other things.
        const rotateCenter = CenterOf(move.movementPattern.around, move.startPosition.setOffset, move.startPosition.lineOffset);
        const hands = move.movementPattern.byHand
          ? new Map<Hand, animation.HandAnimation>([
            [move.movementPattern.byHand, { kind: "Center" }],
            [move.movementPattern.byHand.opposite(), { kind: "None" }],
          ])
          : new Map<Hand, animation.HandAnimation>([
            [Hand.Left, { kind: "None" }],
            [Hand.Right, { kind: "None" }],
          ]);
        return [
          new animation.TransitionAnimationSegment({
            actualAnimation: new animation.RotationAnimationSegment(move.beats,
              startSetPosition,
              endSetPosition,
              move.movementPattern.minAmount,
              {
                center: rotateCenter,
                width: setWidth / 5,
                height: setHeight / 5,
              },
              animation.RotationAnimationFacing.Center,
              {
                transitionBeats: 1,
                minDistance: setWidth,
              },
              hands
            ),
            flags: {
              hands: true,
              rotation: true
            },
            startTransitionBeats: 1
          }),
        ];
      case SemanticAnimationKind.Swing:
        if (move.startPosition.kind !== PositionKind.Circle || move.endPosition.kind !== PositionKind.Circle) {
          throw "Swing must start in a circle: " + JSON.stringify(move);
        }
        const rotateSwingCenter = CenterOf(move.movementPattern.around, move.startPosition.setOffset, move.startPosition.lineOffset);
        const dancerDistance = move.movementPattern.swingRole === DanceRole.Lark ? DancerDistance.SwingLark : DancerDistance.SwingRobin;

        const swingStart = SemanticToSetPosition({
          ...move.startPosition,
          dancerDistance,
          balance: undefined,
        });
        const beforeUnfold = SemanticToSetPosition({
          ...move.endPosition,
          dancerDistance,
        });
        const unfolded = SemanticToSetPosition({
          ...move.endPosition,
          dancerDistance: DancerDistance.Normal,
          hands: move.movementPattern.swingRole === DanceRole.Lark
            ? new Map<Hand, HandConnection>([[Hand.Right, {hand: Hand.Left, to: HandTo.DancerRight}]])
            : new Map<Hand, HandConnection>([[Hand.Left, {hand: Hand.Right, to: HandTo.DancerLeft}]]),
        });

        return [
          new animation.LinearAnimationSegment(1,
            startSetPosition,
            swingStart,
          ),
          new animation.RotationAnimationSegment(move.beats - 3,
            swingStart,
            beforeUnfold,
            move.movementPattern.minAmount, // TODO base minAmount on number of beats?
            {
              center: rotateSwingCenter,
              width: setWidth / 5,
              height: setHeight / 5,
            },
            animation.RotationAnimationFacing.CenterRelative,
          ),
          new animation.LinearAnimationSegment(1,
            beforeUnfold,
            unfolded,
          ),
          new animation.LinearAnimationSegment(1,
            unfolded,
            endSetPosition,
          ),
        ];
      case SemanticAnimationKind.TwirlSwap:
        const twirlCenter =
          CenterOf(move.movementPattern.around, move.startPosition.setOffset, move.startPosition.lineOffset);
        const aroundTopOrBottom = move.movementPattern.around === CircleSide.Top || move.movementPattern.around === CircleSide.Bottom;
        const postTwirlPos =
        {
          ...endSetPosition,
          leftArmEnd: startSetPosition.leftArmEnd === undefined ? undefined : endSetPosition.leftArmEnd,
          rightArmEnd: startSetPosition.rightArmEnd === undefined ? undefined : endSetPosition.rightArmEnd,
        };
        return [
          new animation.RotationAnimationSegment(move.beats - 1,
            startSetPosition,
            {
              ...postTwirlPos,
              rotation: endSetPosition.rotation - (move.startPosition.hands?.has(Hand.Left) ? 360 : 0)
            },
            180,
            {
              center: twirlCenter,
              width: aroundTopOrBottom ? setWidth : setWidth / 4,
              height: aroundTopOrBottom ? setHeight / 4 : setHeight,
            },
            animation.RotationAnimationFacing.Linear,
            undefined,
            new Map<Hand, animation.HandAnimation>([[[...move.startPosition.hands!.keys()][0], {kind: "Center"}]]),
          ),
          new animation.LinearAnimationSegment(1,
            postTwirlPos,
            endSetPosition)
        ];
    }

    throw "Unsupported move in animateLowLevelMove: " + JSON.stringify(move);
  }

  return [
    animation.LinearAnimationSegment.standStill(startSetPosition),
    ...handleMove(),
    animation.LinearAnimationSegment.standStill(endSetPosition),
  ];
}

export function animateFromLowLevelMoves(moves: Map<DancerIdentity, LowLevelMove[]>): animation.Animation {
  const res = animation.newAnimation();
  for (const [id, moveList] of moves.entries()) {
    res.set(id, moveList.flatMap(move => animateLowLevelMove(move)))
  }
  return res;
}