import * as animation from "./animation.js";
import * as interpreter from "./interpreter.js";
import * as renderer from "./renderer.js";
import { DancerIdentity } from "./danceCommon.js";
import { Move } from "./libfigureMapper.js";
import { animateLowLevelMove, LowLevelMove } from "./lowLevelMove.js";

const body = document.querySelector('body')!;

const wrapperDiv = document.createElement('div');
const canvasDiv = document.createElement('div');
canvasDiv.id = 'canvasDiv';

const canvas = document.createElement('canvas');
for (const str of ["left", "right", "top", "bottom"]) {
  const hallLabel = document.createElement('span');
  hallLabel.innerText = str;
  hallLabel.classList.add(str);
  hallLabel.classList.add('hallLabel');
  canvasDiv.appendChild(hallLabel);
}
canvasDiv.appendChild(canvas);
wrapperDiv.appendChild(canvasDiv);
body.appendChild(wrapperDiv);

const beatSliderLabel = document.createElement('label');
beatSliderLabel.innerText = 'Beat: ';
const beatSlider = document.createElement('input');
beatSlider.type = 'range';
beatSlider.min = '0';
beatSlider.max = '64';
beatSlider.step = 'any';
beatSlider.value = '0';
beatSliderLabel.appendChild(beatSlider);

const beatDisplay = document.createElement('span');
beatDisplay.className = 'beatDisplay';
beatDisplay.innerText = '0';

const ctx = canvas.getContext('2d')!;

// TODO Redo this on resize?
const w = canvas.offsetWidth;
const h = canvas.offsetHeight;
const s = window.devicePixelRatio;

canvas.width = w * s;
canvas.height = h * s;

ctx.translate(canvas.width / 2.0, canvas.height / 2.0);
const scale = 10;
ctx.scale(canvas.height / scale, -canvas.height / scale);
const r = new renderer.Renderer(canvas, ctx);
r.extraLines = 3;
r.extraSets = 3;
r.clear();
r.animation = interpreter.interpretedAnimation;
r.drawSetsWithTrails(0);

const bpmSelector = document.createElement('input');
bpmSelector.type = 'number';
bpmSelector.value = '180';
bpmSelector.id = 'bpm';
bpmSelector.style.width = '4em';
const bpmLabel = document.createElement('label');
bpmLabel.innerText = 'BPM';
bpmLabel.htmlFor = 'bpm';
const playButton = document.createElement('button');
playButton.innerText = "Play";
const numBeats = Math.max(...[...r.animation.values()].map(el => el.reduce((a, s) => a + s.beats, 0)));
beatSlider.max = numBeats.toString();
wrapperDiv.appendChild(bpmSelector);
wrapperDiv.appendChild(bpmLabel);
wrapperDiv.appendChild(playButton);
wrapperDiv.appendChild(document.createElement('br'));
wrapperDiv.appendChild(beatSliderLabel);
wrapperDiv.appendChild(beatDisplay);

beatSlider.addEventListener('input', (ev) => {
  r.drawSetsWithTrails(parseFloat(beatSlider.value));
  beatDisplay.innerText = beatSlider.value;
  restartAnimation(false);
});

let cancelAnim : number | undefined = undefined;
function restartAnimation(startIfPaused: boolean) {
  if (cancelAnim !== undefined) {
    cancelAnimationFrame(cancelAnim);
  } else if (!startIfPaused) {
    return;
  }
  const bpm = parseFloat(bpmSelector.value);
  if (bpm < 0) {
    playAnimation(
      parseFloat(bpmSelector.value),
      beatSlider.value === '0' ? numBeats : parseFloat(beatSlider.value),
      0);
  } else if (bpm > 0) {
    playAnimation(
      parseFloat(bpmSelector.value),
      beatSlider.value === beatSlider.max ? 0 : parseFloat(beatSlider.value),
      numBeats);
  }
}
playButton.addEventListener('click', (ev) => {
  if (cancelAnim !== undefined) {
    cancelAnimationFrame(cancelAnim);
    cancelAnim = undefined;
    playButton.innerText = 'Play';
  } else {
    restartAnimation(true);
  }
});
bpmSelector.addEventListener('change', (ev) => {
  restartAnimation(false);
})
function playAnimation(bpm: number, start: number, end: number) {
  const startTime = Date.now();
  const msPerBeat = (60 * 1000) / bpm;

  function anim() {
    const now = Date.now();
    const msElapsed = now - startTime;
    const beat = start + msElapsed / msPerBeat;
    if (bpm > 0 && beat > end || bpm < 0 && beat < end) {
      beatSlider.value = end.toString();
      beatDisplay.innerText = end.toString();
      r.drawSetsWithTrails(end);
      cancelAnim = undefined;
      playButton.innerText = 'Play';
    } else {
      beatSlider.value = beat.toString();
      beatDisplay.innerText = beat.toString();
      r.drawSetsWithTrails(beat);
      cancelAnim = requestAnimationFrame(anim);
      playButton.innerText = 'Pause';
    }
  }
  anim();
}

function createJsonCell(content: any, rowSpan?: number, id?: DancerIdentity) {
  const cell = document.createElement('td');
  const pre = document.createElement('pre');

  // from https://stackoverflow.com/a/56150320
  function replacer(key, value) {
    if (value instanceof Map) {
      let res = {};
      for (var [k, v] of value.entries()) {
        res[k === undefined ? "undefined" : k.toString()] = v;
      }
      return res;
    } else if (!value) {
      return value;
    } else if (value.enumValue) {
      return value.enumValue;
    } else {
      return value;
    }
  }

  pre.innerText = content ? JSON.stringify(content, replacer, 2) : "";
  cell.appendChild(pre);
  if (rowSpan) {
    cell.rowSpan = rowSpan;
  }
  if (id) {
    cell.classList.add(id.coupleRole.toString());
    cell.classList.add(id.danceRole.toString());
  }
  return cell;
}
function createHeaderCell(content: string, id?: DancerIdentity) {
  const cell = document.createElement('th');
  cell.innerText = content;
  if (id) {
    cell.classList.add(id.coupleRole.toString());
    cell.classList.add(id.danceRole.toString());
    cell.colSpan = 2;
  }
  return cell;
}

const table = document.createElement('table');

const headerRow = document.createElement('tr');
const roles = [DancerIdentity.OnesLark, DancerIdentity.OnesRobin,
               DancerIdentity.TwosLark, DancerIdentity.TwosRobin];
headerRow.appendChild(createHeaderCell("Move"));
for (const role of roles) {
  headerRow.appendChild(createHeaderCell(role.toString(), role));
}
table.appendChild(headerRow);

const byMove: { move: Move, byRole: Map<DancerIdentity, { lowLevelMove?: LowLevelMove, animationSegment: animation.AnimationSegment, numSegments?: number }[]> }[] = [];
for (const [role, moveList] of interpreter.interpretedDance.entries()) {
  for (const move of moveList) {
    let entry = byMove.find(el => el.move === move.move);
    if (!entry) {
      entry = { move: move.move, byRole: new Map<DancerIdentity, { lowLevelMove?: LowLevelMove, animationSegment: animation.AnimationSegment, numSegments?: number }[]>() };
      byMove.push(entry);
    }
    let forRole = entry.byRole.get(role);
    if (!forRole) {
      forRole = [];
      entry.byRole.set(role, forRole);
    }
    const animation = animateLowLevelMove(move);
    let first = true;
    for (const animationSegment of animation) {
      if (first) {
        forRole.push({ lowLevelMove: move, animationSegment, numSegments: animation.length  });
        first = false;
      } else {
        forRole.push({ animationSegment });
      }
    }
  }
}

for (const infoForMove of byMove) {
  const moveRow = document.createElement('tr');
  const numRows = Math.max(...[...infoForMove.byRole.values()].map(l => l.length));
  moveRow.appendChild(createJsonCell(infoForMove.move, numRows));

  for (let i = 0; i < numRows; i++) {
    const row = i == 0 ? moveRow : document.createElement('tr');

    for (const role of roles) {
      const lowLevelMove = infoForMove.byRole.get(role)?.at(i);
      if (!lowLevelMove) {
        row.appendChild(createJsonCell(undefined, undefined, role));
      } else if (lowLevelMove.lowLevelMove) {
        row.appendChild(createJsonCell(lowLevelMove.lowLevelMove, lowLevelMove.numSegments, role));
      }
      row.appendChild(createJsonCell(lowLevelMove?.animationSegment, undefined, role));
    }

    table.appendChild(row);
  }
}

body.appendChild(table);
